package cache

import (
	"errors"
	"time"

	"github.com/jellydator/ttlcache/v3"
)

type SimpleCache[K comparable, V any] interface {
	Add(key K, value V) error
	AddWithTTL(key K, value V, ttl time.Duration) error
	Get(key K) (V, error)
	GetWithLoader(key K, loader func(key K) (V, time.Duration, error)) (V, error)
	Keys() []K
}

type Options struct {
	SizeLimit  uint64
	DefaultTTL time.Duration
}

func NewSimpleCache[K comparable, V any](options ...Options) SimpleCache[K, V] {
	opts := []ttlcache.Option[K, V]{
		ttlcache.WithDisableTouchOnHit[K, V](),
	}
	if len(options) > 0 {
		o := options[0]
		if o.SizeLimit > 0 {
			opts = append(opts, ttlcache.WithCapacity[K, V](o.SizeLimit))
		}
		if o.DefaultTTL > 0 {
			opts = append(opts, ttlcache.WithTTL[K, V](o.DefaultTTL))
		}
	}

	c := ttlcache.New[K, V](opts...)
	return &simpleCache[K, V]{
		data: c,
	}
}

type simpleCache[K comparable, V any] struct {
	data *ttlcache.Cache[K, V]
}

func (c *simpleCache[K, V]) Add(key K, value V) error {
	return c.AddWithTTL(key, value, ttlcache.DefaultTTL)
}

func (c *simpleCache[K, V]) AddWithTTL(key K, value V, ttl time.Duration) error {
	item := c.data.Set(key, value, ttl)
	if item == nil {
		return errors.New("failed to add item")
	}
	return nil
}

func (c *simpleCache[K, V]) Get(key K) (V, error) {
	item := c.data.Get(key)
	if item == nil {
		var zero V
		return zero, errors.New("item not found")
	}
	return item.Value(), nil
}

func (c *simpleCache[K, V]) GetWithLoader(key K, loader func(key K) (V, time.Duration, error)) (V, error) {
	loaderWrapper := ttlcache.LoaderFunc[K, V](
		func(t *ttlcache.Cache[K, V], key K) *ttlcache.Item[K, V] {
			value, ttl, err := loader(key)
			if err != nil {
				return nil
			}
			return t.Set(key, value, ttl)
		},
	)
	item := c.data.Get(key, ttlcache.WithLoader[K, V](loaderWrapper))
	if item == nil {
		var zero V
		return zero, errors.New("item not found")
	}
	return item.Value(), nil
}

func (c *simpleCache[K, V]) Keys() []K {
	return c.data.Keys()
}
